/*  indentify archive type from name magic
 *  initialize device i/o as needed and
 *  set up function pointers to service
 *  routines
 */
#include <stdio.h>
#include <string.h>
#include "tar.h"
#include "devices.h"

void name_magic(char *name);
void errfcn(int fd, long offset, int mode);
int nullfcn(int fd,int mode); /* for faking setmode */
int fakeopen(char *,int, int); /* for faking setmode */
int fakecreat(char *,int); /* for faking setmode */

int (*openptr)(char *, int, int) = open;
int (*creatptr)(char *, int) = creat;
int (*readptr)(int, char *, int) = read;
int (*writeptr)(int, char *, int) = write;
int (*closeptr)( int ) = close;
int (*setmodeptr)( int,int ) = setmode;
long (*lseekptr)(int, long, int) = lseek;

int
rmtopen (char *name, int mode, int perm)
{
	name_magic(name);
	return ((*openptr)(name,mode, perm));
}
int
rmtcreat(char *name, int mode)
{
	name_magic(name);
	return ((*creatptr)(name, mode));
}

int
rmtread	(int fd, char *buf, int size) 
{
	(void) kbhit();  /* ^C/break check */
	return  ((*readptr)(fd, buf, size));
}

int
rmtwrite(int fd, char *buf, int size)
{
	(void) kbhit();  /* ^C/break check */
	return ((*writeptr)(fd,buf, size));
}

int
rmtclose( int fd )
{
	return ((*closeptr)(fd));
}

int
rmtlseek(int fd, long offset, int mode)
{
	return ((*lseekptr)(fd, offset, mode));
}

int
rmtsetmode( int fd,int mode)
{
return ((*setmodeptr)( fd,mode));
}

void
errfcn(int fd, long offset, int mode)
{
	fprintf(stderr,"Unsupported function called, exiting\n");
	exit (1);
}
int
nullfcn(int fd,int mode)  /* for faking setmode */
{
	return (0);
}

int
fakeopen(char *name,int mode, int perms)
{
	return (__REM_BIAS +1);
} 

int fakecreat(char *name,int mode)
{
	return (__REM_BIAS +1);
} 

#if 0
/* probably need these too... */
#define rmtaccess	access
#define rmtstat		stat
#define rmtlstat	lstat
#define rmtioctl	ioctl
#define rmtdup		dup
#define rmtfstat	fstat
#define rmtfcntl	fcntl
#define rmtisatty	isatty
#endif

/* do minimal checking on name, let specific routines do final parsing */
void
name_magic(char *name)
{
	char *cp;
	int count = 0;
	if (strcmp(name,"-") == 0) {
		return; /* stdin/out already set */
	}
	if (strncmp(name,"/dev/ct",7) == 0 ||
	    strncmp(name,"/dev/nrct",9) == 0 ) { /* aspi device */
		openptr = aspiopen;
		creatptr = aspicreat;
		readptr = aspiread;
		writeptr = aspiwrite;
		closeptr = aspiclose;
		lseekptr = aspilseek;
		setmodeptr = nullfcn;
		return;
	}
        cp = name;
	while (*cp) {
		if (*cp++ == ':' ) {
			count++;
		}
	}
#ifdef TCP
	if (strncmp(name,"/dev/rshd/",9) == 0 ) {  /* wait for remote tcp connect */
		openptr =  do_rshd;
		creatptr = do_rshd;
		readptr =  tcpread;
		writeptr = tcpwrite;
		closeptr = tcpclose;
		lseekptr = errfcn;
		setmodeptr = nullfcn;
		return;
	}
	if (strncmp(name,"/dev/rsh/",8) == 0 ||
	    strncmp(name,"/dev/rexec/",10) == 0  ||
	    count > 2 ) {  /* remote tcp connect */
		openptr =  tcpopen;
		creatptr = tcpcreat;
		readptr =  tcpread;
		writeptr = tcpwrite;
		closeptr = tcpclose;
		lseekptr = errfcn;
		setmodeptr = nullfcn;
		return;
	}
#endif
#ifdef DIO
	if (strncmp(name,"a:dio",5) == 0 ||
	    strncmp(name,"b:dio",5) == 0 ) { /* raw floppy device */
		openptr =  dio_open3;
		creatptr = dio_open2;
		readptr =  dio_read;
		writeptr = dio_write;
		closeptr = dio_close;
		lseekptr = dio_lseek;
		setmodeptr = nullfcn;
		return;
	}
#endif

}

#ifdef TCP
void set_dos_devices(void) {

		openptr =  open;
		creatptr = creat;
		readptr =  read;
		writeptr = write;
		closeptr = close;
		lseekptr = lseek;
		setmodeptr = setmode;
}

void set_rsh_devices(void) {

		openptr =  fakeopen;
		creatptr = fakecreat;
		readptr =  tcpread;
		writeptr = tcpwrite;
		closeptr = tcpclose;
		lseekptr = errfcn;
		setmodeptr = nullfcn;
}
#endif