//
// This is a Windows Encapsulation for GNU TAR for DOS.
// GNU TAR for DOS is written by:
//
//      First version, 1992-1993, known as RTAR or Pisa TAR,
//         Luigi Rizzo, F. Rosellini & G. Mercaldo, University of Pisa
//      Later versions, 1993-1994, known as GNU TAR for DOS
//         Les Mikesell, ???
//      Windows Encapsulation, 1994, known as GNU TAR for Windows
//         Will van Geest, University of Delft (W.J.M.vGeest@ET.TUDelft.NL)
//
// This program is free software; you can redistribute it and/or modify
// it, but you may not sell it.
//
// This program is distributed in the hope that it will be useful,
// but without any warranty; without even the implied warranty of
// merchantability or fitness for a particular purpose.
//
// GNU TAR, GNU TAR for DOS and this Encapsulation for GNU TAR for DOS
// are protected by the GNU GENERAL PUBLIC LICENSE, see the file COPYING.
//
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <dir.h>
#include "winsock.h"
#include "tar.h"
#include "devices.h"

#define APIENTRY  PASCAL
#define WNDPROC   FARPROC

// Windows Sockets version
#define WSVERSION 0x101

// messages for windows sockets returns
#define WM_STDINOUT           (WM_USER + 1)
#define WM_ERRINOUT           (WM_USER + 2)
#define WM_UDPINOUT           (WM_USER + 3)

// An way to block without using blocking calls and other utilities
#define WAIT_UNTIL(x) while (!(x) && GetMessage(&WinMsg, NULL, 0, 0)) { TranslateMessage(&WinMsg); DispatchMessage(&WinMsg); } if (!(x)) exit(TE_ERROR)
#define dim(x) (sizeof(x) / sizeof(x[0]))

// Connection modes
#define DAEMON 1
#define TRIGGER 2

// Error numbers
#define TE_ERROR        1
#define TE_NOSERVICE    2
#define TE_NOWAIT       3
#define TE_NOBIND       4
#define TE_NOLISTEN     5
#define TE_NOCREATE     6
#define TE_NOACCEPT     7
#define TE_NORECV       8
#define TE_LIMIT        9
#define TE_NOSEND      10
#define TE_NOPORT      11
#define TE_NOPEER      12
#define TE_NOCONNECT   13
#define TE_NOSETOPT    14
#define TE_QUIT        15
#define TE_CLSCONN     16
#define TE_NOLOCAL     17
#define TE_NOLOGIN     18
#define TE_NOGETOPT    19
#define TE_NOPERM      20
#define TE_WAITFAIL    21
#define TE_NOMALLOC    22
#define TE_NOCMD       23

// Some typedefs
typedef struct {                    // associates messages (or menu ids)
   UINT Code;                       // with handler functions
   LONG (*Fxn)(HWND, UINT, UINT, LONG);
} DECODEWORD;

typedef struct {                    // associates an error code with text
   UINT err;
   int repWSAcode;
   char *sztext;
} ERRENTRY;

typedef struct {                    // buffers, pointers etc. for socket operations
   SOCKET sock;
   int connect;
   int instr;
   int ridx;
   int rchars;
   int rfull;
   int wfree;
   char string[128];
   char rbuf[512];
} SockBuf;

// Functions in this file
int  APIENTRY WinMain(HANDLE hInstance, HANDLE hPrevInstance, LPSTR lpCmdLine, int nCmdShow);
int  GetPrivProfInt(char *Item, int Default, char *File);
char *GetPrivProfStr(char *Item, char *Default, char *File);
void ExitCleanUp(void);
BOOL InitApp(HANDLE hInstance);
BOOL InitInstance(HANDLE hINSTANCE, int nCmdShow);
int  do_rshd(char *name, int mode, int perm);
void do_rsh(char *service, char *remhost, char *remuser, char* locuser, char *cmd);
void do_quote(void);
LONG FAR APIENTRY NetWndProc(HWND hWnd, UINT wMsg, UINT wParam, LONG lParam);
LONG DoStdInOut(HWND hWnd, UINT wMsg, UINT wParam, LONG lParam);
LONG DoErrInOut(HWND hWnd, UINT wMsg, UINT wParam, LONG lParam);
LONG DoUdpInOut(HWND hWnd, UINT wMsg, UINT wParam, LONG lParam);
void CreateErrSocket(HWND hWnd, int port);
void ReportWSError(UINT err);
void ReportError(UINT Err);
char *ReadStr(SockBuf *p);
int  check_cmd(char *line);

int  tcpread(int fd, unsigned char *rbuffer, int size);
int  tcpwrite(int fd, unsigned char *rbuffer, int blockdim);
int  tcpcreat(char *name, int mode);
int  tcpopen(char *name, int mode, int perm);
int  tcpclose(int fd);         
void check_ar_file(char *devname);

void check_rsh_device(void);
char **get_rsh_cmd(int *argc);
int  system(const char *cmd);
int  spawnl(int mode, char *path, char *arg0, ...);
int  execl(char *path, char *arg0, ...);
void harderr(void (*fnc)(int));
void hardresume(int mode);

char *GetDefArFile(void);
void FakeExit(int status);

// Functions in other files
char **parse_option(char *input,int *pargc,char **av);
void main(int argc, char **argv);
char *basename(char *path);
void set_rsh_devices(void);

// Variables in other files
extern POINT _ScreenSize;
extern POINT _WindowSize;
extern char  _WindowTitle[];

// Global variables
DECODEWORD netMsgs[] = {               // network window messages & handlers
   WM_STDINOUT,         DoStdInOut,
   WM_ERRINOUT,         DoErrInOut,
   WM_UDPINOUT,         DoUdpInOut
};

ERRENTRY wsErrs[] = {                  // error text for windows sockets errors
   WSAVERNOTSUPPORTED, 0, "This version of Windows Sockets is not supported",
   WSASYSNOTREADY,     0, "Windows Sockets is not present or is not responding",
};

ERRENTRY finErrs[] = {                 // GNU TAR specific error text
   TE_NOSERVICE, 1, "Cannot locate port for service",
   TE_NOWAIT,    1, "Asyncselect call failed",
   TE_NOBIND,    1, "Bind failed",
   TE_NOLISTEN,  1, "Listen failed",
   TE_NOCREATE,  1, "Cannot create socket",
   TE_NOACCEPT,  1, "Accept failed",
   TE_NORECV,    1, "Error occurred during retrieval",
   TE_LIMIT,     0, "String too long",
   TE_NOSEND,    1, "Error occured during sending",
   TE_NOPORT,    0, "No free port for error stream",
   TE_NOPEER,    1, "Unable to retreive peer address",
   TE_NOCONNECT, 1, "Unable to connect to remote host",
   TE_NOSETOPT,  1, "Setting socket options failed",
   TE_QUIT,      0, "Kill signal from yonder",
   TE_CLSCONN,   0, "Remote host closed connection",
   TE_NOLOCAL,   1, "Unable to retreive local host name",
   TE_NOLOGIN,   0, "Login refused by remote host",
   TE_NOGETOPT,  1, "Unable to get socket options",
   TE_NOPERM,    0, "Permission denied",
   TE_WAITFAIL,  1, "Wait for Winsock call completion aborted",
   TE_NOMALLOC,  0, "Call to allocate memory failed",
   TE_NOCMD,     0, "Invalid command from remote host"
};

char     szAppName[] = "WTAR";         // application's name
HANDLE   hInst;                        // this instance of WTAR
WSADATA  WSAData;                      // windows sockets info return
SOCKET   ListenSock;                   // connects to remote server's socket
HWND     hNetWnd;                      // network window handle
SockBuf  StdSock, ErrSock, UdpSock;    // network buffers
MSG      WinMsg;                       // holds current message
int      Connected;                    // Type of established connection
int      WSAnotStarted = TRUE;         // to close winsock if started
int      ExitStatus;                   // saved parameter of exit function
int      NetSndSize;                   // Size of network buffer for sending
int      DisableNagle;                 // Disable Nagle algorithm for send coalescing
char *   Servers;                      // Servers which are allowed to access us
char *   People;                       // People who are allowed to access us
char *   DefDevice;                    // Default device
char     UdpReply[128];                // String to reply to UDP query
int  *   a;                            // Pointer to new argument count
char **  av = 0;                       // Pointer to array of new arguments
char *   Status;                       // String to send back to quote requestor
char *   AfterHost;                    // Pointer after static string for quote reply
char *   Rservice;                     // Pointers to parameters in /dev/rxxxx/ specs
char *   Rhostname;
char *   Rremuser;
char *   Rlocuser;
char *   Rcmd;
//
// Standard windows functions
//

//
// WinMain -- windows calls this to start the application.
//
#pragma argsused
int APIENTRY WinMain(HANDLE hInstance, HANDLE hPrevInstance,
    LPSTR lpCmdLine, int nCmdShow) {

   char buf[256];
   char *pExt;
   char **argv;
   int argc;

   hInst = hInstance;                        // save our instance handle

   // Make some settings for the EasyWin window
   strcpy(_WindowTitle, "GNU TAR for Windows");
   _ScreenSize.x = 80;
   _ScreenSize.y = 100;
   _WindowSize.x = (GetSystemMetrics(SM_CXSCREEN) * 2) / 3;
   _WindowSize.y = GetSystemMetrics(SM_CYSCREEN) / 2;

   if (!hPrevInstance) {                     // if first instance,
      if (!InitApp(hInstance)) {             // register window classes
         MessageBox(hNetWnd, "Can't initialize WTAR", szAppName, MB_ICONSTOP | MB_OK);
         return(FALSE);
      }
   }

   if (!InitInstance(hInstance, nCmdShow)) { // per instance initialization
      MessageBox(hNetWnd, "Can't initialize WTAR", szAppName, MB_ICONSTOP | MB_OK);
      return(FALSE);
   }

   // Get location of wtar.ini file. Look first in directory of executable
   GetModuleFileName(hInstance, buf, sizeof(buf));
   if ((pExt = strrchr(buf, '\\')) != NULL) {
      strcpy(++pExt, "WTAR.INI");
      if (access(buf, 0) == -1) {
         buf[0] = '\0';
      } // ini file not in executable directory
   } else {
      buf[0] = '\0';
   } // no .EXE extension?
               
   // If no ini file in excutable directory, search path
   if ((buf[0] == '\0') && ((pExt = searchpath("WTAR.INI")) != NULL)) {
      strcpy(buf, pExt);
   }

   // Read settings from .ini file if present
   NetSndSize = GetPrivProfInt("SendSize", 0, buf);
   DisableNagle = GetPrivProfInt("DisableNagle", 1, buf);
   Servers = GetPrivProfStr("Servers", "", buf);
   People = GetPrivProfStr("People", "", buf);
   DefDevice = GetPrivProfStr("DefaultDevice", "tar.out", buf);

   atexit(ExitCleanUp);
   argc = 0;
   argv = NULL;
   memset(buf, 0, sizeof(buf));
   strcpy(buf, szAppName);
   strcat(buf, " ");
   if (lpCmdLine[0] != '\0') {
      strcat(buf, lpCmdLine);
   } else {
      printf("Give GNU TAR arguments: ");
      if(gets(&buf[strlen(buf)]) == NULL) exit(0);
   }   
   argv = parse_option(buf, &argc, argv);
   main(argc, argv);

   while (GetMessage(&WinMsg, NULL, 0, 0))   // loop til WM_QUIT
   {
      TranslateMessage(&WinMsg);
      DispatchMessage(&WinMsg);
   }

   if (!WSAnotStarted) WSACleanup();         // disconnect from winsock
   return(WinMsg.wParam);                    // return to windows
}

int GetPrivProfInt(char *Item, int Default, char *File) {

   int tmp;

   // Get int from file if file name given
   if (File[0] != '\0') {
      tmp = GetPrivateProfileInt("GNU TAR", Item, Default, File);
   } else {
      tmp = Default;
   }

   return(tmp);
}

char *GetPrivProfStr(char *Item, char *Default, char *File) {

   char buf[1024];
   char *tmp;

   // Get string from file if file name given
   if (File[0] != '\0') {
      GetPrivateProfileString("GNU TAR", Item, Default, buf, sizeof(buf), File);
   } else {
      strcpy(buf, Default);
   }

   // Allocate the memory
   if ((tmp = malloc(strlen(buf)+1)) == NULL) {
      ReportError(TE_NOMALLOC);
   }

   // Copy the data
   strcpy(tmp, buf);
   return(tmp);
}

void ExitCleanUp(void) {

   // Empty the message queue
   PostQuitMessage(0);                       // Make sure there's a WM_QUIT
   while (GetMessage(&WinMsg, NULL, 0, 0)) {
      TranslateMessage(&WinMsg);
      DispatchMessage(&WinMsg);
   }

   // Free allocated memory
   free(Servers);
   free(People);
   free(DefDevice);

   // Disconnect from winsock
   if (!WSAnotStarted) {
      closesocket(StdSock.sock);                // Nicely close all sockets
      closesocket(ErrSock.sock);                // if open?
      closesocket(UdpSock.sock);
      closesocket(ListenSock);
      WSACleanup();
   }
   if (!f_keep_window && (ExitStatus == 0)) PostQuitMessage(0); // Destroy the EasyWin window
   return;
}

//
// InitApp -- initialization for all instances of application.
// Registers main window class.
//
BOOL InitApp(HANDLE hInstance)
{
   WNDCLASS    wndclass;

   wndclass.style         = 0;
   wndclass.lpfnWndProc   = NetWndProc;
   wndclass.cbClsExtra    = 0;
   wndclass.cbWndExtra    = 0;
   wndclass.hInstance     = hInstance;
   wndclass.hIcon         = 0;
   wndclass.hCursor       = 0;
   wndclass.hbrBackground = 0;
   wndclass.lpszMenuName  = NULL;
   wndclass.lpszClassName = szAppName;

   return(RegisterClass(&wndclass));
}

//
// InitInstance -- initializes this instance of app, and creates windows.
//
#pragma argsused
BOOL InitInstance(HANDLE hInstance, int nCmdShow) {

   HWND hEasyWnd;
   HICON hTarIcon;

   hNetWnd = CreateWindow( szAppName, szAppName,
               WS_OVERLAPPEDWINDOW | WS_VSCROLL,
               CW_USEDEFAULT, CW_USEDEFAULT,
               CW_USEDEFAULT, CW_USEDEFAULT,
               NULL, NULL, hInstance, NULL);

   if (!hNetWnd) return(FALSE);
   _InitEasyWin();
   if ((hEasyWnd = FindWindow("BCEasyWin", "GNU TAR for Windows")) != NULL) {
      hTarIcon = LoadIcon(hInstance, "WTARIcon");
      SetClassWord(hEasyWnd, GCW_HICON, hTarIcon);
   }
   return(TRUE);
}
 
//
// do_rshd -- called by tar module to start the rsh daemon. 
//
#pragma argsused
int do_rshd(char *name, int mode, int perm) {

   struct servent FAR *pServent;
   SOCKADDR_IN sin;
   int tmp;
   char *pStr;
   char RemUs[18];

   // Get the activation mode
   mode = (name != NULL) && (name[0] == '\0');

   // We're busy
   Status = " is starting daemon/trigger mode";
   // Start the winsock dll if not done yet 
   if (WSAnotStarted) {
      if ((tmp = WSAStartup(WSVERSION, &WSAData)) != 0) { // register task with winsock
         ReportWSError(tmp);            
         DestroyWindow(hNetWnd);               // kill application window &
      } //WSAStartup failed                    // signal app exit
      WSAnotStarted = FALSE;
      do_quote();
   } //WSA not started yet

   // No connections yet
   Connected = 0;

   // Get port for service
   if ((pServent = getservbyname("shell", "tcp")) == NULL) {
      ReportError(TE_NOSERVICE);
   } //getservbyname failed

   // Create a socket
   if ((ListenSock = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
      ReportError(TE_NOCREATE);
   } //socket failed

   // Bind to the service port
   memset(&sin, 0, sizeof(sin));
   sin.sin_family = AF_INET;
   sin.sin_port = pServent->s_port;
   sin.sin_addr.s_addr = INADDR_ANY;
   if (bind(ListenSock, (LPSOCKADDR) &sin, sizeof(sin)) == SOCKET_ERROR) {
      ReportError(TE_NOBIND);
   } //bind failed

   // Async select
   if (WSAAsyncSelect(ListenSock, hNetWnd, WM_STDINOUT, FD_ACCEPT) == SOCKET_ERROR) {
      ReportError(TE_NOWAIT);
      return(FALSE);
   } // WSAAsyncSelect failed

   // Start listening
   StdSock.connect = FALSE;                // No connection yet
   if (listen(ListenSock, 1) == SOCKET_ERROR) {
      ReportError(TE_NOLISTEN);
   } //listen failed

   // Now we wait for an incoming call
   if (mode) {
      Status = " is in daemon mode";      // Started with --rshd cq --wait
      printf("Entering daemon mode, waiting for incoming request.\n");
   } else {
      Status = " is in trigger mode";     // Started with /dev/rshd
      printf("Entering trigger mode, waiting for incoming request.\n");
   }
   WAIT_UNTIL(StdSock.connect);

   tmp = sizeof(sin);
   if ((StdSock.sock = accept(ListenSock, (LPSOCKADDR) &sin, (int FAR *) &tmp)) == INVALID_SOCKET) {
      ReportError(TE_NOACCEPT);
   } // accept failed

   // Stop listening
   closesocket(ListenSock);

   // Log incoming request
   tmp = sizeof(sin);
   if (getpeername(StdSock.sock, (LPSOCKADDR) &sin, (int FAR *) &tmp) == SOCKET_ERROR) {
      ReportError(TE_NOPEER);
   } // getpeername failed  
   printf("Request from %Fs received\n", inet_ntoa(sin.sin_addr));

   // Get size of transmit buffer. This is tricky, because Trumpet Winsock has
   // some problem with its buffer size reports
   if (NetSndSize <= 0) { 
      if ((NetSndSize == 0) && (strstr(WSAData.szDescription, "Trumpet") != NULL)) {
         NetSndSize = 1024;
      } else {
         tmp = sizeof(NetSndSize);
         if (getsockopt(StdSock.sock, SOL_SOCKET, SO_SNDBUF, (char *)&NetSndSize, &tmp) == SOCKET_ERROR) {
            ReportError(TE_NOGETOPT);
         } // getsockopt failed
      } // optimal NetSndSize found
   } // NetSndSize not valid

   // Disable Nagle
   tmp = DisableNagle;
   if (setsockopt(StdSock.sock, IPPROTO_TCP, TCP_NODELAY, (char *)&tmp, sizeof(tmp)) == SOCKET_ERROR) {
      ReportError(TE_NOSETOPT);
   } // setsockopt failed

   // Async select
   if (WSAAsyncSelect(StdSock.sock, hNetWnd, WM_STDINOUT, FD_READ | FD_WRITE | FD_CLOSE) == SOCKET_ERROR) {
      ReportError(TE_NOWAIT);
      return(FALSE);
   } // WSAAsyncSelect failed

   // Writing is free as a start
   StdSock.wfree = TRUE;

   // Get port for error stream and make a connection
   tmp = atoi(ReadStr(&StdSock));
   CreateErrSocket(hNetWnd, tmp);

   // Get remote username and remember it
   strcpy(RemUs, ReadStr(&StdSock));
   printf("The remote user is %s\n", RemUs);

   // Get local username and forget it
   pStr = ReadStr(&StdSock);

   // Get command
   pStr = ReadStr(&StdSock);
   printf("The command is %s%s\n", pStr, (mode) ? "" : " (ignored)");

   // Verify host and user
   if (((*Servers != 0) && (strstr(Servers, inet_ntoa(sin.sin_addr)) == NULL)) ||
      ((*People != 0) && (strstr(People, RemUs) == NULL))) {
      tcpwrite(0, "\1Permission denied\n", -1);
      ReportError(TE_NOPERM);
   }

   // Decode the command, if in real daemon mode
   if (mode) {
      pStr[strlen(pStr)+1] = '\0';        // End in double zero for check_cmd
      if (check_cmd(pStr)) {
         tcpwrite(0, "\1Can't make pipe\n", -1);
         ReportError(TE_NOCMD);
      } // Not a tar command
   } else {
      f_reblock++;                        // in trigger mode, blocks will come in bits & pieces
   }

   // Send acknowledge
   tcpwrite(0, "\0", 1);

   // Everything ok
   Status = " is engaged";
   Connected = DAEMON;
   return(__REM_BIAS+1);
}

//
// do_rsh -- called by tar to send a command to a remote host
//
void do_rsh(char *service, char *remhost, char *remuser, char* locuser, char *cmd) {

   static int alport = -1;
   struct servent FAR *pServent;
   struct hostent FAR *pHostent;
   SOCKADDR_IN sin;
   int tmp;
   char CmdStr[128 + 2 + 16 + 16];

   // We're busy
   Status = " is starting a rexec/rsh session";

   // Start the winsock dll if not done yet 
   if (WSAnotStarted) {
      if ((tmp = WSAStartup(WSVERSION, &WSAData)) != 0) { // register task with winsock
         ReportWSError(tmp);            
         DestroyWindow(hNetWnd);               // kill application window &
      } //WSAStartup failed                    // signal app exit
      WSAnotStarted = FALSE;
      do_quote();
   } //WSA not started yet

   // No connections yet
   Connected = 0;

   // Get port for service
   if ((pServent = getservbyname(service, "tcp")) == NULL) {
      ReportError(TE_NOSERVICE);
   } //getservbyname failed

   // Create a socket
   if ((StdSock.sock = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
      ReportError(TE_NOCREATE);
   } //socket failed

   // Randomly select a starting source port number
   if ((int) alport == -1) {
      randomize();
      alport = rand() % (IPPORT_RESERVED/2) + (IPPORT_RESERVED/2);
   } // alport not set yet

   // Bind the socket to a free privileged port
   memset(&sin, 0, sizeof(sin));
   sin.sin_family = AF_INET;
   sin.sin_addr.s_addr = 0;
   while (1) {

      sin.sin_port = htons(alport);
      if (bind(StdSock.sock, (LPSOCKADDR) &sin, sizeof(sin)) == SOCKET_ERROR) {
         if (WSAGetLastError() != WSAEADDRINUSE) {
            ReportError(TE_NOBIND);       // Failure to bind to a port
         } // Other error than port in use
      } else {
         break;                           // bind succeeded, leave loop
      } // bind failed

      // Try another port
      if (--alport == (IPPORT_RESERVED/2)) {
         ReportError(TE_NOPORT);          // All privileged ports tested
      } // all ports in use
   } // find free port
   if (--alport == (IPPORT_RESERVED/2)) alport = IPPORT_RESERVED - 1;

   // Get remote host address, try inet_addr first, then gethostbyname
   memset(&sin, 0, sizeof(sin));
   sin.sin_family = AF_INET;
   sin.sin_port = pServent->s_port;
   if ((sin.sin_addr.s_addr = inet_addr(remhost)) == INADDR_NONE) {
      if ((pHostent = gethostbyname(remhost)) == NULL) {
         ReportError(TE_NOPEER);
      } // gethostbyname failed
      sin.sin_family = pHostent->h_addrtype;
      memcpy(&sin.sin_addr, pHostent->h_addr, pHostent->h_length);
   } // inet_addr failed

   // Start AsyncSelect
   if (WSAAsyncSelect(StdSock.sock, hNetWnd, WM_STDINOUT, FD_CONNECT | FD_READ | FD_WRITE | FD_CLOSE) == SOCKET_ERROR) {
      ReportError(TE_NOWAIT);
   } // WSAAsyncSelect failed

   // Writing is free now
   StdSock.wfree = TRUE;

   // Connect to remote host
   StdSock.connect = TRUE;
   if (connect(StdSock.sock, (LPSOCKADDR) &sin, sizeof(sin)) == SOCKET_ERROR) {
      if (WSAGetLastError() == WSAEWOULDBLOCK) {
         StdSock.connect = FALSE;
      } else {
         ReportError(TE_NOCONNECT);
      } // WASEWOULDBLOCK
   } // connect failed
   WAIT_UNTIL(StdSock.connect);

   // Get size of transmit buffer. This is tricky, because Trumpet Winsock has
   // some problem with its buffer size reports
   if (NetSndSize <= 0) { 
      if ((NetSndSize == 0) && (strstr(WSAData.szDescription, "Trumpet") != NULL)) {
         NetSndSize = 1024;
      } else {
         tmp = sizeof(NetSndSize);
         if (getsockopt(StdSock.sock, SOL_SOCKET, SO_SNDBUF, (char *)&NetSndSize, &tmp) == SOCKET_ERROR) {
            ReportError(TE_NOGETOPT);
         } // getsockopt failed
      } // optimal NetSndSize found
   } // NetSndSize not valid

   // Disable Nagle
   tmp = DisableNagle;
   if (setsockopt(StdSock.sock, IPPROTO_TCP, TCP_NODELAY, (char *)&tmp, sizeof(tmp)) == SOCKET_ERROR) {
      ReportError(TE_NOSETOPT);
   } // setsockopt failed

   // Connect succeeded, build complete string to send to remote
   tmp = 0;                            // Start at the beginning
   strcpy(&CmdStr[tmp], "0");          // Port number for error stream
   tmp += strlen("0") + 1;             // Next starting point (include trailing zero)

   strncpy(&CmdStr[tmp], remuser, 16); // Append remote username
   CmdStr[tmp + 15] = '\0';            // Make sure maximum 15 chars
   tmp += strlen(&CmdStr[tmp]) + 1;    // Next starting point (include trailing zero)

   strncpy(&CmdStr[tmp], locuser, 16); // Append local username
   CmdStr[tmp + 15] = '\0';            // Make sure maximum 15 chars
   tmp += strlen(&CmdStr[tmp]) + 1;    // Next starting point (include trailing zero)

   strncpy(&CmdStr[tmp], cmd, 128);    // Append the command to execute
   CmdStr[tmp + 127] = '\0';           // Make sure maximum 127 chars
   tmp += strlen(&CmdStr[tmp]) + 1;    // Include trailing zero 

   // Send command to the remote host
   tcpwrite(0, CmdStr, tmp);

   // Wait for acknowledge and verify acceptance
   printf("Connected to %Fs\n", inet_ntoa(sin.sin_addr));
   tcpread(StdSock.sock, StdSock.rbuf, sizeof(StdSock.rbuf));
   if (StdSock.rbuf[0] != '\0') {
      printf("%s\n", &StdSock.rbuf[1]);
      exit(TE_NOLOGIN);
   }

   // Everything ok
   Connected = TRIGGER;
   Status = " has started a remote session";
   return;
}

//
// do_quote -- start an UDP listener so a remote host can verify we're alive
//
#pragma argsused
void do_quote(void) {

   struct servent *pServent;
   SOCKET s;
   SOCKADDR_IN sin;
   int tmp;

   // Start the winsock dll if not done yet 
   if (WSAnotStarted) {
      if ((tmp = WSAStartup(WSVERSION, &WSAData)) != 0) { // register task with winsock
         ReportWSError(tmp);            
         DestroyWindow(hNetWnd);               // kill application window &
      } //WSAStartup failed                    // signal app exit
      WSAnotStarted = FALSE;
   } //WSA not started yet

   // Get port for service
   if ((pServent = getservbyname("quote", "udp")) == NULL) {
      ReportError(TE_NOSERVICE);
   } //getservbyname failed

   // Create a socket
   if ((s = socket(AF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET) {
      ReportError(TE_NOCREATE);
   } //socket failed

   // Bind to an port
   memset(&sin, 0, sizeof(sin));
   sin.sin_family = AF_INET;
   sin.sin_port = pServent->s_port;
   sin.sin_addr.s_addr = INADDR_ANY;
   if (bind(s, (LPSOCKADDR) &sin, sizeof(sin)) == SOCKET_ERROR) {
      ReportError(TE_NOBIND);
   } //bind failed

   //Get hostname to use in quote to reply
   if (gethostname(UdpReply, sizeof(UdpReply)) == SOCKET_ERROR) {
      ReportError(TE_NOLOCAL);
   } // gethostname failed

   // Remember where the hostname ends
   if ((AfterHost = strchr(UdpReply, '.')) == NULL) {
      AfterHost = strchr(UdpReply, '\0');
   }

   // Writing is not yet allowed
   UdpSock.wfree = FALSE;

   // Socket operational, Async select for data received
   UdpSock.sock = s;
   if (WSAAsyncSelect(UdpSock.sock, hNetWnd, WM_UDPINOUT, FD_READ | FD_WRITE) == SOCKET_ERROR) {
      ReportError(TE_NOWAIT);
   } // WSAAsyncSelect failed

   // Everything ok
   return;
}

//
//  Message handlers
//

//
// NetWndProc -- callback function for network window.  The network window
// is a child window created especially to receive windows sockets network
// messages.  The function decodes and routes to appropriate message handler.
//
LONG FAR APIENTRY NetWndProc(HWND hWnd, UINT wMsg, UINT wParam, LONG lParam)
{
   int i;

   for (i = 0; i < dim(netMsgs); i++) {
      if (wMsg == netMsgs[i].Code) return((*netMsgs[i].Fxn)(hWnd, wMsg, wParam, lParam));
   }

   return(DefWindowProc(hWnd, wMsg, wParam, lParam));
}

//
// Read strings from stdin/stdout port.
//
#pragma argsused
LONG DoStdInOut(HWND hWnd, UINT wMsg, UINT wParam, LONG lParam)
{

   if (WSAGETSELECTERROR(lParam)) {
      WSASetLastError(WSAGETSELECTERROR(lParam));     // something went wrong
      ReportError(TE_WAITFAIL);
      return(FALSE);
   } // Some error

   switch (WSAGETSELECTEVENT(lParam)) {
      case FD_ACCEPT:
         StdSock.connect = TRUE;
         return(FALSE);

      case FD_CONNECT:
         StdSock.connect = TRUE;
         return(FALSE);

      case FD_READ:
         StdSock.rfull = TRUE;
         return(FALSE);

      case FD_WRITE:
         StdSock.wfree = TRUE;
         return(FALSE);

      case FD_CLOSE:
         Connected = 0;
         printf("Closing all sockets\n");
         return(FALSE);
   }

   // Everything done
   return(FALSE);
}

#pragma argsused
LONG DoErrInOut(HWND hWnd, UINT wMsg, UINT wParam, LONG lParam) {

   int rchars;

   if (WSAGETSELECTERROR(lParam)) {
      WSASetLastError(WSAGETSELECTERROR(lParam));     // something went wrong
      ReportError(TE_WAITFAIL);
      return(FALSE);
   }

   switch (WSAGETSELECTEVENT(lParam)) {
      case FD_CONNECT:
         ErrSock.connect = TRUE;
         return(FALSE);

      case FD_READ:
         if (((rchars = recv(ErrSock.sock, ErrSock.rbuf, sizeof(ErrSock.rbuf), 0)) == SOCKET_ERROR) &&
            (WSAGetLastError() != WSAEWOULDBLOCK)) {
            ReportError(TE_NORECV);
         } // recv failed
         if (rchars == 0) {
            closesocket(ErrSock.sock);
         } // connection closed by remote side
         if ((rchars > 0) && (ErrSock.rbuf[0] == '\2')) {
            ReportError(TE_QUIT);
         } // QUIT signal received
         return(FALSE);

      case FD_WRITE: 
         ErrSock.wfree = TRUE;
         return(FALSE);
                             
   }

   // Everything done
   return(FALSE);
}

#pragma argsused
LONG DoUdpInOut(HWND hWnd, UINT wMsg, UINT wParam, LONG lParam) {

   int rchars;
   static SOCKADDR_IN sin;
   static int sinlen;

   if (WSAGETSELECTERROR(lParam)) {
      WSASetLastError(WSAGETSELECTERROR(lParam));     // something went wrong
      ReportError(TE_WAITFAIL);
      return(FALSE);
   }                                       

   if (WSAIsBlocking()) return(FALSE);

   switch (WSAGETSELECTEVENT(lParam)) {
      case FD_READ:
         sinlen = sizeof(sin);
         if (((rchars = recvfrom(UdpSock.sock, UdpSock.rbuf, sizeof(UdpSock.rbuf), 0, (LPSOCKADDR) &sin, &sinlen)) == SOCKET_ERROR) &&
            (WSAGetLastError() != WSAEWOULDBLOCK)) {
            ReportError(TE_NORECV);
         } // recv failed 
         UdpSock.wfree = (rchars != 0);
      case FD_WRITE:
         if (UdpSock.wfree) {
            strcpy(AfterHost, Status);
            // Verify host and user
            if (((*Servers != 0) && (strstr(Servers, inet_ntoa(sin.sin_addr)) == NULL)) ||
               ((*People != 0) && (strstr(People, UdpSock.rbuf) == NULL))) {
               strcat(AfterHost, ", permission denied");
            }
            if ((sendto(UdpSock.sock, UdpReply, strlen(UdpReply), 0, (LPSOCKADDR) &sin, sinlen) == SOCKET_ERROR) &&
               (WSAGetLastError() != WSAEWOULDBLOCK)) {
               ReportError(TE_NOSEND);
            } else {
               UdpSock.wfree = FALSE;
            } // send failed
         }
   } 

   // Everything done
   return(FALSE);
}

//
// Utility functions for do_rshd, do_rsh and do_quote
//

void CreateErrSocket(HWND hWnd, int port) {

SOCKADDR_IN sin;
static u_short alport = -1;  // To circumvent EADDRINUSE because TIME_WAIT
int tmp;

   if (port != 0) {
      if ((int) alport == -1) {
         randomize();
         alport = rand() % (IPPORT_RESERVED/2) + (IPPORT_RESERVED/2);
      } // alport not set yet

      // Create socket
      if ((ErrSock.sock = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
         ReportError(TE_NOCREATE);
      } // socket failed

      // Bind socket to a free privileged port
      memset(&sin, 0, sizeof(sin));
      sin.sin_family = AF_INET;
      sin.sin_addr.s_addr = 0;
      while (1) {

         sin.sin_port = htons(alport);
         if (bind(ErrSock.sock, (LPSOCKADDR) &sin, sizeof(sin)) == SOCKET_ERROR) {
            if (WSAGetLastError() != WSAEADDRINUSE) {

               // Cannot bind to a privileged port
               ReportError(TE_NOBIND);
            } // Other error than port in use
         } else {
            break; // bind succeeded, leave loop
         } // bind failed

         // Try another port
         if (--alport == (IPPORT_RESERVED/2)) {

            // Cannot bind to a privileged port
            ReportError(TE_NOPORT);
         } // all ports in use
      } // find free port
      if (--alport == (IPPORT_RESERVED/2)) alport = IPPORT_RESERVED - 1;
         
      // Get address of peer
      memset(&sin, 0, sizeof(sin));
      tmp = sizeof(sin);
      if (getpeername(StdSock.sock, (LPSOCKADDR) &sin, (int FAR *) &tmp) == SOCKET_ERROR) {
         ReportError(TE_NOPEER);
      } // getpeername failed

      // Set port number in address of peer
      sin.sin_port = htons(port);

      // Async select for data received
      if (WSAAsyncSelect(ErrSock.sock, hWnd, WM_ERRINOUT, FD_CONNECT | FD_READ | FD_WRITE) == SOCKET_ERROR) {
         ReportError(TE_NOWAIT);
      } // WSAAsyncSelect failed

      // Connect to remote host
      ErrSock.connect = TRUE;
      if (connect(ErrSock.sock, (LPSOCKADDR) &sin, sizeof(sin)) == SOCKET_ERROR) {
         if (WSAGetLastError() == WSAEWOULDBLOCK) {
            ErrSock.connect = FALSE;
         } else {
            ReportError(TE_NOCONNECT);
         } // WASEWOULDBLOCK
      } // connect failed
      WAIT_UNTIL(ErrSock.connect);

   } // valid port number

   // Everything done
   return;
}
 
//
// ReportWSError -- prompt user with a windows sockets error message.
//
void ReportWSError(UINT Err) {

   int i;

   for (i = 0; i < dim(wsErrs); i++) {
      if (Err == wsErrs[i].err) {
         printf("%s\n", wsErrs[i].sztext);
         return;
      }
   }

   printf("Windows Sockets reports error %d", Err);
}

//
// ReportError -- prompt user with a TCP specific error
//
void ReportError(UINT Err) {

   int i, tmp, size;
   char buf[81];

   for (i = 0; i < dim(finErrs); i++) {
      if (Err == finErrs[i].err) {
         if (finErrs[i].repWSAcode) {
            tmp = WSAGetLastError();
            if (tmp == 0) tmp = WSABASEERR;
            if (tmp >= WSABASEERR) {
               size=LoadString(hInst, tmp, buf, 80);
            } else {
               size=0;
            }
            if (size==0) sprintf(buf, "[%i] Invalid WINSOCK error code!", tmp);
            printf("%s,\n%s\n", finErrs[i].sztext, buf);
         } else {
            printf("%s.\n", finErrs[i].sztext);
         }
         exit(Err);
      }
   }

   printf("Unknown error,\nWindows Sockets error %d\n", WSAGetLastError());
   exit(Err);
}

//
// Read a null terminated string from the stdin/out socket. 
//
char *ReadStr(SockBuf *p) {

   int rchars;

   // Copy byte by byte until null char or until p->instr over limit
   do {
        
      // Check limit
      if (p->instr > sizeof(p->string)) {
         ReportError(TE_LIMIT);
      } // limit

      // Check number of chars in buffer
      if (p->ridx >= p->rchars) {

         // Get a new buffer
         p->ridx = p->rchars = 0;

         // Get one good buffer
         rchars = SOCKET_ERROR;
         while (rchars == SOCKET_ERROR) {
            WAIT_UNTIL(StdSock.rfull);       // Wait until message received

            // Get the buffer from the network
            if ((rchars = recv(p->sock, p->rbuf, sizeof(p->rbuf), 0)) == SOCKET_ERROR) {
               if (WSAGetLastError() == WSAEWOULDBLOCK) {
                  p->rfull = FALSE;          // No net buffers free, wait until next time
               } else {
                  ReportError(TE_NORECV);    // Some unexpected error
               } //WSAEWOULDBLOCK
            } // recv failed
         } // good buffer received

         // Check if connection still alive
         if (rchars == 0) {
            ReportError(TE_CLSCONN);
         } // connection dead

         p->rchars = rchars;
      } //buffer empty

      // Copy the byte
      p->string[p->instr] = p->rbuf[p->ridx];
      p->ridx++;
   } while (p->string[p->instr++] != 0);

   // The next call returns a new string
   p->instr = 0;
   return(p->string);
}

int check_cmd(char *line) {
   char *cp,*cp2,*cmd, *tmp;
   char was;
   int tar;

   cp=line;
   while (isspace(*cp)) cp++ ;
   cp2=cp+1 ;
   while (*cp2 && !isspace(*cp2)) cp2++;
   was = *cp2;
   *cp2 = '\0';
   cmd = basename(cp);
   tar=(((tmp = strrchr(cmd, 't')) != NULL) && (strcmp(tmp,"tar") == 0));
   *cp2 = was;
   if (tar) {
      av = parse_option(line,a,av);
   }
   return (!tar);
}

//
// Functions called by GNU TAR modules to manipulate remote archives
//

#pragma argsused
int tcpread(int fd, unsigned char *rbuffer, int size) {

   int rchars;

   // Get one good buffer
   rchars = SOCKET_ERROR;
   while (rchars == SOCKET_ERROR) {
      WAIT_UNTIL(StdSock.rfull);       // Wait until message received

      // Get the buffer from the network
      if ((rchars = recv(StdSock.sock, rbuffer, size, 0)) == SOCKET_ERROR) {
         if (WSAGetLastError() == WSAEWOULDBLOCK) {
            StdSock.rfull = FALSE;     // No net buffers free, wait until next time
         } else {
            ReportError(TE_NORECV);    // Some unexpected error
         } //WSAEWOULDBLOCK
      } // recv failed

      // Check if connection still alive
      if (rchars == 0) {
         ReportError(TE_CLSCONN);
      } // connection dead
   } // buffer received 

   // Return how much we did get
   return (rchars);
}

#pragma argsused
int tcpwrite(int fd, unsigned char *rbuffer, int blockdim) {

   int totsent;
   int count;
   int size;

   // Use string length if blockdim is -1
   if (blockdim == -1) blockdim = strlen(rbuffer);

   // Send all data
   totsent = 0;
   size = (blockdim < NetSndSize) ? blockdim : NetSndSize;
   while (totsent < blockdim) {
      WAIT_UNTIL(StdSock.wfree);          // Wait until net buffer free

      // Write one buffer
      if ((count = send(StdSock.sock, &rbuffer[totsent], size, 0)) == SOCKET_ERROR) {
         if (WSAGetLastError() == WSAEWOULDBLOCK) {
            StdSock.wfree = FALSE;        // No net buffers free, wait until next time
         } else {
           ReportError(TE_NOSEND);        // Some unexpected error
         } //WSAEWOULDBLOCK
      } else {
         // Remember what has been sent and what has to be sent next;
         totsent += count;
         if (totsent > blockdim - NetSndSize) size = blockdim - totsent;
      } // send failed
   } // data left to send

   // Return how much we did send.
   return (totsent);
}

#pragma argsused
int tcpcreat(char *name, int mode) {

   check_ar_file(name);
   do_rsh(Rservice, Rhostname, Rremuser, Rlocuser, Rcmd);
   return(1);
}

#pragma argsused
int tcpopen(char *name, int mode, int perm) {

   check_ar_file(name);
   do_rsh(Rservice, Rhostname, Rremuser, Rlocuser, Rcmd);
   f_reblock++;                           // Blocks are arriving in bits and pieces
   return(1);
}

#pragma argsused
int tcpclose(int fd) {

   closesocket(StdSock.sock);
   closesocket(ErrSock.sock);
   closesocket(UdpSock.sock);
   closesocket(ListenSock);
   return(0);
}

void check_ar_file(char *devname) {

   int f_remote = 0;
    
   if (strncmp(devname,"/dev/rsh/",9) == 0) {
      f_remote++;
      Rservice = "shell";
      devname += 9;
   } else if (strncmp(devname,"/dev/rexec/",11) == 0) {
      f_remote++;
      Rservice = "exec";
      devname += 11;
   }

   // Get all items
   Rhostname = strtok(devname,":");
   Rremuser = strtok(NULL, ":");
   Rlocuser = strtok(NULL, ":");
   Rcmd = strtok(NULL, "\4");

   // Now get the items in place
   if (Rcmd == NULL) {
      Rcmd = Rlocuser;
      Rlocuser = Rremuser;
   }

   // Now all items should be there
   if ((Rhostname == NULL) || (Rremuser == NULL) || (Rlocuser == NULL) || (Rcmd == NULL)) {
      printf("Syntax error in remote command\n");
      exit(EX_ARGSBAD);
   } // one or more items missing;

   // Final check
   if (!f_remote) {
      printf("How did we get here?\n");
      exit(EX_SYSTEM);
   } 
}

//
// Other functions called by GNU TAR modules
//

void check_rsh_device(void) {

   if (Connected == DAEMON) {
      set_rsh_devices();
   }
}

/* get connection, parse remote command, adjust argc, return new argv */
char **get_rsh_cmd(int *argc) {
   a = argc; /* pointer to caller's var */
   *a = 0;   /* restart args */
   av = 0; 
   do_rshd("",0,0);
   return (av);
}

//
// Replacements for MSDOS functions called by GNU TAR modules
//
 
int system( const char FAR *cmd) {

   char buf[256];
   char *cmdcom;
   HTASK hTask;

   if ((cmdcom = getenv("COMSPEC")) == NULL) {
      strcpy(buf, "\\command.com");
   } else {
      strcpy(buf, cmdcom);
   }

   if (cmd[0] != '-' || cmd[1] != '\0') {
      strcat(buf, " /c ");
      strncat(buf, cmd, sizeof(buf) - strlen(buf) - 1);
   }
   hTask = WinExec(buf, SW_NORMAL);
   if (hTask > HINSTANCE_ERROR) {
      for (;;) {
         while (PeekMessage(&WinMsg, NULL, 0, 0, PM_REMOVE)) {     // Loop til WM_QUIT
            if (WinMsg.message == WM_QUIT) exit(__REM_BIAS +1);
            TranslateMessage(&WinMsg);
            DispatchMessage(&WinMsg);
         }
         if (!GetModuleUsage(hTask)) break;
      }
   }
   return(0);
}

#pragma argsused
int spawnl(int mode, char *path, char *arg0, ...) {

   system("-");
   return(0);
}

#pragma argsused
int execl(char *path, char *arg0, ...) {

   va_list ap;
   char buf[256];
   char *arg;

   va_start(ap, arg0);
   strcpy(buf, arg0);

   while ((arg = (char *) va_arg(ap, char *)) != NULL) {
      strcat(buf, " ");
      strcat(buf, arg);
   }

   va_end(ap);
   WinExec(buf, SW_NORMAL);
   return(0);
}

#pragma argsused
void harderr(void (*fnc)(int)) {

   SetErrorMode(SEM_FAILCRITICALERRORS);
}

#pragma argsused
void hardresume(int mode) {

}

//
// Functions to fake add functionality to library routines
//

//
// Function to get pointer to default file name
// Called by tar main module by defining DEF_AR_FILE as GetDefArFile()
//
char *GetDefArFile(void) {

   return(DefDevice);
} 

//
// Fake exit handler to get hold of the exit status. In ExitCleanUp the exit
// status is used to determine if the log window must stay. The fake handler
// is called by defining exit as FakeExit, e.g. -Dexit=FakeExit
// Undefine fake exit handler, now we really must leave
//
#ifdef exit
#undef exit
#endif
// Prototype the real function
void exit(int exitcode);

void FakeExit(int status) {

   ExitStatus = status;    // Remember the parameter
   exit(status);           // Let exit and ExitCleanUp do the rest
}
